import {
    isNull,
    isUndefined,
    isEqual,
    uniq,
    isEmpty,
    max
} from 'lodash';
import { strings } from './constants';

import { KeyEntities, getSortedKeyList } from '../tree/treeUtil';

function getPosition(level: any, index: any) {
    return `${level}-${index}`;
}

export function isValid(val: any) {
    return !isNull(val) && !isUndefined(val);
}

export function normalizedArr(val: any) {
    if (!Array.isArray(val)) {
        return [val];
    } else {
        return val;
    }
}

/**
 * Traverse all the data by `treeData`.
 */
function traverseDataNodes(treeNodes: any, callback: any) {
    const processNode = (node: any, ind?: any, parent?: any) => {
        const children = node ? node.children : treeNodes;
        let item: any = null;
        // Process node if is not root
        if (node) {
            const key = parent ? getPosition(parent.key, ind) : `${ind}`;
            const { children, ...resetInfo } = node;
            item = {
                data: { ...resetInfo },
                ind,
                key,
                level: parent ? parent.level + 1 : 0,
                parentKey: parent ? parent.key : null,
                path: parent ? [...parent.path, key] : [key],
                valuePath: parent ? [...parent.valuePath, node.value] : [node.value]
            };

            callback(item);
        }

        // Process children node
        if (children) {
            children.forEach((subNode: any, subIndex: any) => {
                processNode(subNode, subIndex, item);
            });
        }
    };

    processNode(null);
}

export function convertDataToEntities(dataNodes: any) {
    const keyEntities: any = {};

    traverseDataNodes(dataNodes, (data: any) => {
        const { key, parentKey } = data;
        const entity = { ...data };

        keyEntities[key] = entity;

        // Fill children
        // entity.parent = keyEntities[parentKey];
        // if (entity.parent) {
        //     entity.parent.children = entity.parent.children || [];
        //     entity.parent.children.push(entity);
        // }
        const entityParent = keyEntities[parentKey];
        if (entityParent) {
            // entityParent.children = entityParent.children || [];
            // entityParent.children.push(entity);
            entityParent.childrenKeys = entityParent.childrenKeys || [];
            entityParent.childrenKeys.push(key);
        }
    });
    return keyEntities;
}

export function findKeysForValues(value: any, keyEntities: any) {
    const valuePath = normalizedArr(value);
    const res = Object.values(keyEntities)
        .filter((item: any) => isEqual(item.valuePath, valuePath))
        .map((item: any) => item.key);
    return res;
}

export function calcMergeType(autoMergeValue: boolean, leafOnly: boolean): string {
    let mergeType: string;
    if (leafOnly) {
        mergeType = strings.LEAF_ONLY_MERGE_TYPE;
    } else if (autoMergeValue) {
        mergeType = strings.AUTO_MERGE_VALUE_MERGE_TYPE;
    } else {
        mergeType = strings.NONE_MERGE_TYPE;
    }
    return mergeType;
}

export function findChildKeys(keys: string[], options: any, omitKeys: any[] = []) {
    const res: any[] = [];
    keys &&
        keys.forEach(key => {
            const opts = options[key];
            // opts &&
            //     opts.children &&
            //     opts.children.forEach((child: any) => {
            //         if (!omitKeys.length || !omitKeys.includes(child.key)) {
            //             res.push(child.key);
            //         }
            //     });
            opts &&
                opts.childrenKeys &&
                opts.childrenKeys.forEach((child: any) => {
                    if (!omitKeys.length || !omitKeys.includes(child.key)) {
                        res.push(child.key);
                    }
                });
        });
    return res;
}

export function findSiblingKeys(selectedKeys: string[], options: any, self = true) {
    const par: any[] = [];
    // selectedKeys.forEach(item => {
    //     if (options[item] && options[item].parent) {
    //         par.push(options[item].parent.key);
    //     }
    // });
    selectedKeys.forEach(item => {
        if (options[item] && options[item].parentKey) {
            par.push(options[item].parentKey);
        }
    });
    

    const res = findChildKeys(uniq(par), options, self ? [] : selectedKeys);
    return res;
}

export function findAncestorKeys(selectedKeys: string[], options: any, self = true) {
    const res: any[] = [];
    // Recursively find the parent element
    // const findPar = (item: any) => {
    //     if (item.parent) {
    //         res.push(item.parent.key);
    //         findPar(item.parent);
    //     }
    // };
    const findPar = (item: any) => {
        if (item.parentKey) {
            res.push(item.parentKey);
            findPar(options[item.parentKey]);
        }
    };
    selectedKeys.forEach(item => {
        options[item] && findPar(options[item]);
        if (self) {
            res.push(item);
        }
    });
    return res;
}

export function calcCheckedKeysForUnchecked(key: string, keyEntities: KeyEntities, checkedKeys: Set<string>, halfCheckedKeys: Set<string>) {
    const descendantKeys = findDescendantKeys([key], keyEntities, true);
    const nodeItem = keyEntities[key];
    descendantKeys.forEach(descendantKey => {
        if (checkedKeys.has(descendantKey)) {
            checkedKeys.delete(descendantKey);
        }
        if (halfCheckedKeys.has(descendantKey)) {
            halfCheckedKeys.delete(descendantKey);
        }
    });
    const calcCurrLevel = (node: any) => {
        // const par = node.parent;
        const par = keyEntities[node.parentKey];
        // no parent
        if (!par) {
            return;
        }
        // Has a parent node, and the parent node is not checked or halfChecked
        if (!checkedKeys.has(par.key) && !halfCheckedKeys.has(par.key)) {
            return;
        }
        // Has a parent node, and the parent node is checked or halfChecked
        // eslint-disable-next-line @typescript-eslint/no-shadow
        const { key } = node;
        const siblingKeys = findSiblingKeys([key], keyEntities);
        // eslint-disable-next-line @typescript-eslint/no-shadow
        const anyChecked = siblingKeys.some(key => checkedKeys.has(key) || halfCheckedKeys.has(key));
        const ancestorKeys = findAncestorKeys([key], keyEntities, false);
        // If there is checked or halfChecked in the sibling node, you need to change the parent node to halfChecked
        if (anyChecked) {
            ancestorKeys.forEach(itemKey => {
                if (checkedKeys.has(itemKey)) {
                    checkedKeys.delete(itemKey);
                    halfCheckedKeys.add(itemKey);
                }
            });
            // If there is no checked or halfChecked in the sibling node, you need to change the parent node to unchecked
        } else {
            if (checkedKeys.has(par.key)) {
                checkedKeys.delete(par.key);
            }
            if (halfCheckedKeys.has(par.key)) {
                halfCheckedKeys.delete(par.key);
            }
            calcCurrLevel(par);
        }
    };
    calcCurrLevel(nodeItem);
    return {
        checkedKeys,
        halfCheckedKeys,
    };
}

export function calcCheckedKeysForChecked(key: string, keyEntities: KeyEntities, checkedKeys: Set<string>, halfCheckedKeys: Set<string>) {
    const descendantKeys = findDescendantKeys([key], keyEntities, true);
    const nodeItem = keyEntities[key];
    checkedKeys = new Set([...checkedKeys, key]);
    const calcCurrLevel = (node: any) => {
        // if (!node.parent) {
        //     return;
        // }
        const par = keyEntities[node.parentKey];
        if (!par) {
            return;
        }
        // eslint-disable-next-line @typescript-eslint/no-shadow
        const { key } = node;
        const siblingKeys = findSiblingKeys([key], keyEntities);
        // eslint-disable-next-line @typescript-eslint/no-shadow
        const allChecked = siblingKeys.every(key => checkedKeys.has(key));
        if (!allChecked) {
            const ancestorKeys = findAncestorKeys([key], keyEntities, false);
            halfCheckedKeys = new Set([...halfCheckedKeys, ...ancestorKeys]);
        } else {
            // const par = node.parent;
            checkedKeys.add(par.key);
            calcCurrLevel(par);
        }
    };
    calcCurrLevel(nodeItem);
    return {
        checkedKeys: new Set([...checkedKeys, ...descendantKeys]),
        halfCheckedKeys,
    };
}

export function calcCheckedKeys(values: any, keyEntities: KeyEntities) {
    const keyList = Array.isArray(values) ? values : [values];
    const descendantKeys = findDescendantKeys(keyList, keyEntities, true);
    /**
     * Recursively find the parent element. Because the incoming nodes are all checked,
     * their descendants must be checked. That is to say, if the descendant nodes have
     *  disabled+unchecked nodes, their ancestor nodes will definitely not be checked
     */
    const checkedKeys = new Set([...descendantKeys]);
    let halfCheckedKeys = new Set([]);
    let visited: any[] = [];

    const levelMap:{[key: number]: string[]} = getSortedKeyList(keyList, keyEntities);

    const calcCurrLevel = (node: any) => {
        // const { key, parent, level } = node;
        const { key, parentKey, level } = node;
        const parent = keyEntities[parentKey];
        // If the node does not have a parent node, or the node has been processed just now, no processing is done
        if (!parent || visited.includes(key)) {
            return;
        }

        const siblingKeys = findSiblingKeys([key], keyEntities);
        // visited for caching to avoid double counting
        visited = [...visited, ...siblingKeys];
        const allChecked = siblingKeys.every((siblingKey: string) => checkedKeys.has(siblingKey));
        if (!allChecked) {
            const ancestorKeys = findAncestorKeys([key], keyEntities, false);
            halfCheckedKeys = new Set([...halfCheckedKeys, ...ancestorKeys]);
        } else {
            checkedKeys.add(parent.key);
            // IMPORTANT! parent level may not exist in original level map; if add to the end directly may destroy the hierarchical order
            if (level - 1 in levelMap && level) {
                levelMap[level - 1].push(parent.key);
            } else {
                levelMap[level - 1] = [parent.key];
            }
        }
    };
    // Loop keyList from deepest Level to topLevel, bottom up
    while (!isEmpty(levelMap)) {
        const maxLevel = max(Object.keys(levelMap).map(key => Number(key)));
        levelMap[maxLevel].forEach((key: string) => calcCurrLevel(keyEntities[key]));
        delete levelMap[maxLevel];
    }

    return {
        checkedKeys,
        halfCheckedKeys,
    };
}

export function findDescendantKeys(selectedKeys: string[], options: KeyEntities, self = true) {
    const res: string[] = [];
    const findChild = (item: any) => {
        if (!item) {
            return;
        }
        // const { children } = item;
        // const hasChildren = isValid(children);
        // if (hasChildren) {
        //     children.forEach((child: any) => {
        //         res.push(child.key);
        //         findChild(options[child.key]);
        //     });
        // }
        const { childrenKeys } = item;
        const hasChildren = isValid(childrenKeys);
        if (hasChildren) {
            childrenKeys.forEach((childKey: any) => {
                res.push(childKey);
                findChild(options[childKey]);
            });
        }
    };
    selectedKeys.forEach(item => {
        if (self) {
            res.push(item);
        }
        findChild(options[item]);
    });
    return res;
}

export function normalizeKeyList(keyList: any, keyEntities: KeyEntities, leafOnly = false) {
    const res: string[] = [];
    const keyListSet = new Set(keyList);
    if (!leafOnly) {
        keyList.forEach((key: string) => {
            if (!keyEntities[key]) {
                return;
            }
            const { parentKey } = keyEntities[key];
            if (parentKey && keyListSet.has(parentKey)) {
                return;
            }
            res.push(key);
        });
    } else {
        keyList.forEach(key => {
            // if (keyEntities[key] && !isValid(keyEntities[key].children)) {
            if (keyEntities[key] && !isValid(keyEntities[key].childrenKeys)) {
                res.push(key);
            }
        });
    }
    return res;
}

export function calcDisabledKeys(keyEntities: KeyEntities) {
    const disabledKeys = Object.keys(keyEntities).filter(key => keyEntities[key].data.disabled);
    const { checkedKeys } = calcCheckedKeys(disabledKeys, keyEntities);
    return checkedKeys;
}